/* Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */

#include "../wireguard.h"
#include "../crypto/blake2b.h"
#include "noise.h"
#include <linux/bitmap.h>

#define COUNTER_BACKTRACK_SIZE (sizeof(unsigned long) * 8)
static inline bool counter_validate(union noise_counter *counter, u64 their_counter)
{
	bool ret = false;
	u64 difference;
	spin_lock_bh(&counter->receive.lock);

	if (unlikely(counter->receive.counter >= REJECT_AFTER_MESSAGES))
		goto out;

	if (likely(their_counter > counter->receive.counter)) {
		difference = their_counter - counter->receive.counter;
		bitmap_shift_left(&counter->receive.backtrack, &counter->receive.backtrack, difference, COUNTER_BACKTRACK_SIZE);
		if (likely(difference <= COUNTER_BACKTRACK_SIZE))
			set_bit(difference - 1, &counter->receive.backtrack);
		counter->receive.counter = their_counter;
		ret = true;
		goto out;
	}

	difference = counter->receive.counter - their_counter;
	if (unlikely(difference == 0 || difference > COUNTER_BACKTRACK_SIZE))
		goto out;
	ret = !test_and_set_bit(difference - 1, &counter->receive.backtrack);

out:
	spin_unlock_bh(&counter->receive.lock);
	return ret;
}

bool noise_message_encrypt(struct noise_message_data *dst, const u8 *src_plaintext, size_t src_len, struct noise_symmetric_key *key)
{
	u64 nonce;

	if (unlikely(!key || !key->is_valid || time_is_before_eq_jiffies(key->birthdate + REJECT_AFTER_TIME)))
		return false;

	nonce = atomic64_inc_return(&key->counter.counter);
	if (nonce >= REJECT_AFTER_MESSAGES) {
		key->is_valid = false;
		return false;
	}
	nonce = cpu_to_le64(nonce);
	dst->counter = nonce;

	return chacha20poly1305_encrypt(dst->encrypted_data, src_plaintext, src_len, NULL, 0, (uint8_t *)&nonce, key->key);
}

bool noise_message_decrypt(u8 *dst_plaintext, const struct noise_message_data *src, size_t src_len, struct noise_symmetric_key *key)
{
	if (unlikely(!key || src_len <= sizeof(struct noise_message_data) || !key->is_valid || time_is_before_eq_jiffies(key->birthdate + REJECT_AFTER_TIME) || key->counter.receive.counter >= REJECT_AFTER_MESSAGES))
		return false;

	if (!chacha20poly1305_decrypt(dst_plaintext, src->encrypted_data, src_len - sizeof(struct noise_message_data), NULL, 0, (uint8_t *)&src->counter, key->key))
		return false;

	return counter_validate(&key->counter, le64_to_cpu(src->counter));
}

/* These functions are sloppy with atomic operations on the nonce
 * and don't take locks because we're relying on the caller,
 * handshake.c, to always be holding the handshake lock when
 * calling these functions. */

bool noise_handshake_encrypt(u8 *dst_ciphertext, const u8 *src_plaintext, size_t src_len, struct noise_symmetric_key *key, u8 hash[NOISE_HASH_LEN])
{
	blake2b_state blake;
	u64 nonce = key->counter.receive.counter;

	if (unlikely(!key->is_valid || time_is_before_eq_jiffies(key->birthdate + REJECT_AFTER_TIME) || nonce >= REJECT_AFTER_MESSAGES)) {
		key->is_valid = false;
		return false;
	}

	nonce = cpu_to_le64(nonce);
	if (!chacha20poly1305_encrypt(dst_ciphertext, src_plaintext, src_len, hash, NOISE_HASH_LEN, (uint8_t *)&nonce, key->key))
		return false;
	++key->counter.receive.counter;

	blake2b_init(&blake, NOISE_HASH_LEN);
	blake2b_update(&blake, hash, NOISE_HASH_LEN);
	blake2b_update(&blake, dst_ciphertext, noise_encrypted_len(src_len));
	blake2b_final(&blake, hash, NOISE_HASH_LEN);

	return true;
}

bool noise_handshake_decrypt(u8 *dst_plaintext, const u8 *src_ciphertext, size_t src_len, struct noise_symmetric_key *key, u8 hash[NOISE_HASH_LEN])
{
	blake2b_state blake;
	u64 nonce = key->counter.receive.counter;

	if (unlikely(!key->is_valid || time_is_before_eq_jiffies(key->birthdate + REJECT_AFTER_TIME) || nonce >= REJECT_AFTER_MESSAGES)) {
		key->is_valid = false;
		return false;
	}

	nonce = cpu_to_le64(nonce);
	if (!chacha20poly1305_decrypt(dst_plaintext, src_ciphertext, src_len, hash, NOISE_HASH_LEN, (uint8_t *)&nonce, key->key))
		return false;
	++key->counter.receive.counter;

	blake2b_init(&blake, NOISE_HASH_LEN);
	blake2b_update(&blake, hash, NOISE_HASH_LEN);
	blake2b_update(&blake, src_ciphertext, src_len);
	blake2b_final(&blake, hash, NOISE_HASH_LEN);

	return true;
}

void noise_handshake_nocrypt(u8 *dst, const u8 *src, size_t src_len, u8 hash[NOISE_HASH_LEN])
{
	blake2b_state blake;
	memcpy(dst, src, src_len);
	blake2b_init(&blake, NOISE_HASH_LEN);
	blake2b_update(&blake, hash, NOISE_HASH_LEN);
	blake2b_update(&blake, src, src_len);
	blake2b_final(&blake, hash, NOISE_HASH_LEN);
}
