/* Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */

#include "../wireguard.h"
#include "noise.h"
#include "../crypto/blake2b.h"
#include "../pubkey-hashtable.h"
#include <crypto/algapi.h>

/* <- s
 * ******
 * -> e, dhes, s, dhss, t
 * <- e, dhee, dhes
 */

/* This should be 32 bytes = NOISE_HASH_LEN = NOISE_SYMMETRIC_KEY_LEN */
static const u8 protocol_specifier[] =
{ 'N', 'o', 'i', 's', 'e', ' ', 'W', 'i',
  'r', 'e', 'G', 'u', 'a', 'r', 'd', ' ',
  'z', 'x', '2', 'c', '4', ' ', '2', '0',
  '1', '5', '-', '1', '0', '-', '1', '4' };

static inline void handshake_init(struct noise_symmetric_key *key, u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN], u8 hash[NOISE_HASH_LEN], const u8 remote_static[NOISE_PUBLIC_KEY_LEN])
{
	blake2b_state blake;
	blake2b_init(&blake, NOISE_HASH_LEN);
	blake2b_update(&blake, protocol_specifier, NOISE_HASH_LEN);
	blake2b_update(&blake, remote_static, NOISE_PUBLIC_KEY_LEN);
	blake2b_final(&blake, hash, NOISE_HASH_LEN);

	memset(key, 0, sizeof(struct noise_symmetric_key));
	spin_lock_init(&key->counter.receive.lock);

	memcpy(chaining_key, protocol_specifier, NOISE_SYMMETRIC_KEY_LEN);
}

static inline void tai64n_now_packed(u8 output[NOISE_TIMESTAMP_LEN])
{
	struct timeval now;
	do_gettimeofday(&now);
	/* http://cr.yp.to/libtai/tai64.html */
	*(u64 *)output = cpu_to_be64(now.tv_sec);
	*(u32 *)(output + sizeof(u64)) = cpu_to_be32(1000 * now.tv_usec + 500);
}

bool noise_handshake_create_initiation(struct noise_message_handshake_initiation *dst, struct noise_peer *peer, u16 self_index)
{
	u8 timestamp[NOISE_TIMESTAMP_LEN];
	bool ret = false;

	down_write(&peer->handshake.lock);

	dst->header.type = NOISE_MESSAGE_HANDSHAKE_INITIATION;
	dst->sender_index = cpu_to_le16(self_index);

	handshake_init(&peer->handshake.key, peer->handshake.chaining_key, peer->handshake.hash, peer->handshake.remote_static);

	/* e */
	curve25519_generate_secret(peer->handshake.ephemeral_private);
	curve25519_generate_public(peer->handshake.ephemeral_public, peer->handshake.ephemeral_private);
	noise_handshake_nocrypt(dst->unencrypted_ephemeral, peer->handshake.ephemeral_public, NOISE_PUBLIC_KEY_LEN, peer->handshake.hash);

	/* dhes */
	if (!noise_mix_dh(&peer->handshake.key, peer->handshake.chaining_key, peer->handshake.ephemeral_private, peer->handshake.remote_static))
		goto out;

	/* s */
	if (!noise_handshake_encrypt(dst->encrypted_static, peer->handshake.static_public, NOISE_PUBLIC_KEY_LEN, &peer->handshake.key, peer->handshake.hash))
		goto out;

	/* dhss */
	if (!noise_mix_dh(&peer->handshake.key, peer->handshake.chaining_key, peer->handshake.static_private, peer->handshake.remote_static))
		goto out;

	/* t */
	tai64n_now_packed(timestamp);
	if (!noise_handshake_encrypt(dst->encrypted_timestamp, timestamp, NOISE_TIMESTAMP_LEN, &peer->handshake.key, peer->handshake.hash))
		goto out;

	ret = true;

out:
	up_write(&peer->handshake.lock);
	return ret;
}

struct wireguard_peer *noise_handshake_consume_initiation(struct noise_message_handshake_initiation *src, struct wireguard_device *wg)
{
	bool replay_attack;
	size_t i;
	u8 s[NOISE_PUBLIC_KEY_LEN];
	u8 e[NOISE_PUBLIC_KEY_LEN];
	u8 t[NOISE_TIMESTAMP_LEN];
	u8 static_public[NOISE_PUBLIC_KEY_LEN];
	u8 static_private[NOISE_PUBLIC_KEY_LEN];
	struct noise_peer *peer = NULL;
	struct wireguard_peer *wg_peer = NULL;
	struct noise_symmetric_key key;
	u8 hash[NOISE_HASH_LEN];
	u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN];

	/* We use the first peer we can find, for determining this device's public key */
	for (i = 0; i < max_peers; ++i) {
		wg_peer = peer_lookup_by_index(wg, i);
		if (wg_peer)
			break;
	}
	if (!wg_peer)
		goto out;
	down_read(&wg_peer->noise.handshake.lock);
	memcpy(static_public, wg_peer->noise.handshake.static_public, NOISE_PUBLIC_KEY_LEN);
	memcpy(static_private, wg_peer->noise.handshake.static_private, NOISE_PUBLIC_KEY_LEN);
	up_read(&wg_peer->noise.handshake.lock);
	peer_put(wg_peer);
	wg_peer = NULL;

	handshake_init(&key, chaining_key, hash, static_public);

	/* e */
	noise_handshake_nocrypt(e, src->unencrypted_ephemeral, sizeof(src->unencrypted_ephemeral), hash);

	/* dhes */
	if (!noise_mix_dh(&key, chaining_key, static_private, e))
		goto out;

	/* s */
	if (!noise_handshake_decrypt(s, src->encrypted_static, sizeof(src->encrypted_static), &key, hash))
		goto out;

	/* dhss */
	if (!noise_mix_dh(&key, chaining_key, static_private, s))
		goto out;

	/* t */
	if (!noise_handshake_decrypt(t, src->encrypted_timestamp, sizeof(src->encrypted_timestamp), &key, hash))
		goto out;

	/* Lookup which peer we're actually talking to */
	wg_peer = pubkey_hashtable_lookup(&wg->peer_hashtable, s);
	if (!wg_peer)
		goto out;
	down_read(&wg_peer->noise.handshake.lock);
	peer = &wg_peer->noise;
	replay_attack = memcmp(t, peer->handshake.latest_timestamp, NOISE_TIMESTAMP_LEN) <= 0;
	up_read(&peer->handshake.lock);
	if (replay_attack) {
		peer_put(wg_peer);
		wg_peer = NULL;
		goto out;
	}

	key.index = le16_to_cpu(src->sender_index);

	/* Success! Copy everything to peer */
	down_write(&peer->handshake.lock);
	memcpy(peer->handshake.remote_ephemeral, e, NOISE_PUBLIC_KEY_LEN);
	memcpy(peer->handshake.latest_timestamp, t, NOISE_TIMESTAMP_LEN);
	memcpy(peer->handshake.hash, hash, NOISE_HASH_LEN);
	memcpy(peer->handshake.chaining_key, chaining_key, NOISE_SYMMETRIC_KEY_LEN);
	peer->handshake.key = key;
	up_write(&peer->handshake.lock);

out:
	memzero_explicit(&key, sizeof(struct noise_symmetric_key));
	memzero_explicit(static_private, NOISE_PUBLIC_KEY_LEN);
	memzero_explicit(hash, NOISE_HASH_LEN);
	memzero_explicit(chaining_key, NOISE_SYMMETRIC_KEY_LEN);
	return wg_peer;
}

bool noise_handshake_create_response(struct noise_message_handshake_response *dst, struct noise_peer *peer, u16 self_index)
{
	bool ret = false;
	down_write(&peer->handshake.lock);

	dst->header.type = NOISE_MESSAGE_HANDSHAKE_RESPONSE;
	dst->sender_index = cpu_to_le16(self_index);
	dst->receiver_index = cpu_to_le16(peer->handshake.key.index);

	/* e */
	curve25519_generate_secret(peer->handshake.ephemeral_private);
	curve25519_generate_public(peer->handshake.ephemeral_public, peer->handshake.ephemeral_private);
	if (!noise_handshake_encrypt(dst->encrypted_ephemeral, peer->handshake.ephemeral_public, NOISE_PUBLIC_KEY_LEN, &peer->handshake.key, peer->handshake.hash))
		goto out;

	/* dhee */
	if (!noise_mix_dh(&peer->handshake.key, peer->handshake.chaining_key, peer->handshake.ephemeral_private, peer->handshake.remote_ephemeral))
		goto out;

	/* dhes */
	if (!noise_mix_dh(&peer->handshake.key, peer->handshake.chaining_key, peer->handshake.ephemeral_private, peer->handshake.remote_static))
		goto out;

	if (!noise_handshake_encrypt(dst->encrypted_nothing, NULL, 0, &peer->handshake.key, peer->handshake.hash))
		goto out;

	peer->handshake.is_valid = true;

	ret = true;

out:
	up_write(&peer->handshake.lock);
	return ret;
}

struct wireguard_peer *noise_handshake_consume_response(struct noise_message_handshake_response *src, struct wireguard_device *wg)
{
	struct noise_peer *peer;
	struct wireguard_peer *wg_peer;
	struct noise_symmetric_key key;
	u8 hash[NOISE_HASH_LEN];
	u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN];
	u8 e[NOISE_PUBLIC_KEY_LEN];
	u8 ephemeral_private[NOISE_PUBLIC_KEY_LEN];
	u8 static_private[NOISE_PUBLIC_KEY_LEN];

	wg_peer = peer_lookup_by_index(wg, le16_to_cpu(src->receiver_index));
	if (unlikely(!wg_peer))
		return NULL;
	peer = &wg_peer->noise;

	down_read(&peer->handshake.lock);
	key = peer->handshake.key;
	memcpy(hash, peer->handshake.hash, NOISE_HASH_LEN);
	memcpy(chaining_key, peer->handshake.chaining_key, NOISE_SYMMETRIC_KEY_LEN);
	memcpy(ephemeral_private, peer->handshake.ephemeral_private, NOISE_PUBLIC_KEY_LEN);
	memcpy(static_private, peer->handshake.static_private, NOISE_PUBLIC_KEY_LEN);
	up_read(&peer->handshake.lock);

	/* e */
	if (!noise_handshake_decrypt(e, src->encrypted_ephemeral, sizeof(src->encrypted_ephemeral), &key, hash))
		goto fail;

	/* dhee */
	if (!noise_mix_dh(&key, chaining_key, ephemeral_private, e))
		goto fail;

	/* dhes */
	if (!noise_mix_dh(&key, chaining_key, static_private, e))
		goto fail;

	/* decrypt nothing */
	if (!noise_handshake_decrypt(NULL, src->encrypted_nothing, sizeof(src->encrypted_nothing), &key, hash))
		goto fail;


	/* Success! Copy everything to peer */
	down_write(&peer->handshake.lock);
	memcpy(peer->handshake.remote_ephemeral, e, NOISE_PUBLIC_KEY_LEN);
	memcpy(peer->handshake.hash, hash, NOISE_HASH_LEN);
	memcpy(peer->handshake.chaining_key, chaining_key, NOISE_SYMMETRIC_KEY_LEN);
	peer->handshake.key = key;
	peer->handshake.key.index = le16_to_cpu(src->sender_index);
	peer->handshake.is_valid = true;
	up_write(&peer->handshake.lock);
	goto out;

fail:
	peer_put(wg_peer);
	wg_peer = NULL;

out:
	memzero_explicit(&key, sizeof(struct noise_symmetric_key));
	memzero_explicit(hash, NOISE_HASH_LEN);
	memzero_explicit(chaining_key, NOISE_SYMMETRIC_KEY_LEN);
	memzero_explicit(ephemeral_private, NOISE_PUBLIC_KEY_LEN);
	memzero_explicit(static_private, NOISE_PUBLIC_KEY_LEN);
	return wg_peer;
}

void noise_handshake_clear(struct noise_handshake *handshake)
{
	down_write(&handshake->lock);
	memset(&handshake->ephemeral_public, 0, NOISE_PUBLIC_KEY_LEN);
	memset(&handshake->ephemeral_private, 0, NOISE_PUBLIC_KEY_LEN);
	memset(&handshake->remote_ephemeral, 0, NOISE_PUBLIC_KEY_LEN);
	memset(&handshake->hash, 0, NOISE_HASH_LEN);
	memset(&handshake->chaining_key, 0, NOISE_SYMMETRIC_KEY_LEN);
	memset(&handshake->key, 0, sizeof(struct noise_symmetric_key));
	handshake->is_valid = false;
	up_write(&handshake->lock);
}
