/* Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */

#include "../wireguard.h"
#include "noise.h"
#include "../crypto/blake2b.h"
#include "../crypto/chacha20.h"

/* This implements Hugo Krawczyk's HKDF:
 *  - https://eprint.iacr.org/2010/264.pdf
 *  - https://tools.ietf.org/html/rfc5869
 */
static inline void kdf(u8 first_dst[NOISE_SYMMETRIC_KEY_LEN], u8 second_dst[NOISE_SYMMETRIC_KEY_LEN],
		       const u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN], const u8 *data, size_t data_len)
{
	static const u8 one = 1;
	u8 temp_key[BLAKE2B_OUTBYTES];
	blake2b_hmac(temp_key, data, chaining_key, BLAKE2B_OUTBYTES, data_len, NOISE_SYMMETRIC_KEY_LEN);
	blake2b_hmac(temp_key, &one, temp_key, BLAKE2B_OUTBYTES, 1, BLAKE2B_OUTBYTES);
	memcpy(first_dst, temp_key, NOISE_SYMMETRIC_KEY_LEN);
	memcpy(second_dst, temp_key + NOISE_SYMMETRIC_KEY_LEN, NOISE_SYMMETRIC_KEY_LEN);
	memzero_explicit(temp_key, BLAKE2B_OUTBYTES);
}

static inline void key_is_now_fresh(struct noise_symmetric_key *key)
{
	atomic64_set(&key->counter.counter, 0);
	key->counter.receive.backtrack = 0;
	key->birthdate = jiffies;
	key->is_valid = true;
}

static inline void derive_key(struct noise_symmetric_key *first_dst, struct noise_symmetric_key *second_dst,
			      const struct noise_symmetric_key *key, const u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN])
{
	*first_dst = *key;
	*second_dst = *key;
	kdf(first_dst->key, second_dst->key, chaining_key, NULL, 0);
	key_is_now_fresh(first_dst);
	key_is_now_fresh(second_dst);
}

bool noise_mix_dh(struct noise_symmetric_key *key, u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN], const u8 private[NOISE_PUBLIC_KEY_LEN], const u8 public[NOISE_PUBLIC_KEY_LEN])
{
	u8 dh_calculation[NOISE_PUBLIC_KEY_LEN];
	if (!curve25519(dh_calculation, private, public))
		return false;
	kdf(key->key, chaining_key, chaining_key, dh_calculation, NOISE_PUBLIC_KEY_LEN);
	memzero_explicit(dh_calculation, NOISE_PUBLIC_KEY_LEN);
	key_is_now_fresh(key);
	return true;
}

static void clear_previous_session(struct rcu_head *rcu)
{
	struct noise_session_keys *old_session = container_of(rcu, struct noise_session_keys, rcu);
	if (old_session->previous_receiving) {
		memzero_explicit(old_session->previous_receiving, sizeof(struct noise_symmetric_key));
		kfree(old_session->previous_receiving);
	}
	if (old_session->previous_sending) {
		memzero_explicit(old_session->previous_sending, sizeof(struct noise_symmetric_key));
		kfree(old_session->previous_sending);
	}
	kfree(old_session);
	pr_debug("Old noise session cleared and freed\n");
}

bool noise_handshake_begin_session(struct noise_peer *peer, bool i_am_the_initator)
{
	struct noise_session_keys *session = NULL, *old_session;

	down_read(&peer->handshake.lock);
	if (!peer->handshake.is_valid)
		goto fail;
	session = kzalloc(sizeof(struct noise_session_keys), GFP_KERNEL);
	if (!session)
		goto fail;
	session->sending = kzalloc(sizeof(struct noise_symmetric_key), GFP_KERNEL);
	if (!session->sending)
		goto fail;
	session->receiving = kzalloc(sizeof(struct noise_symmetric_key), GFP_KERNEL);
	if (!session->receiving)
		goto fail;
	session->send_with_previous = !i_am_the_initator;
	session->i_am_the_initator = i_am_the_initator;
	if (i_am_the_initator)
		derive_key(session->receiving, session->sending, &peer->handshake.key, peer->handshake.chaining_key);
	else
		derive_key(session->sending, session->receiving, &peer->handshake.key, peer->handshake.chaining_key);
	up_read(&peer->handshake.lock);

	noise_handshake_clear(&peer->handshake);

	mutex_lock(&peer->key_update_lock);
	old_session = rcu_dereference_protected(peer->session, lockdep_is_held(&peer->key_update_lock));
	if (old_session) {
		session->previous_receiving = old_session->receiving;
		session->previous_sending = old_session->sending;
	}
	rcu_assign_pointer(peer->session, session);
	mutex_unlock(&peer->key_update_lock);

	if (old_session)
		call_rcu(&old_session->rcu, clear_previous_session);

	net_dbg_ratelimited("New noise session created\n");

	return true;

fail:
	up_read(&peer->handshake.lock);
	if (session) {
		if (session->sending)
			memzero_explicit(session->sending, sizeof(struct noise_symmetric_key));
		if (session->receiving)
			memzero_explicit(session->receiving, sizeof(struct noise_symmetric_key));
		kfree(session->sending);
		kfree(session->receiving);
		kfree(session);
	}
	return false;
}
