/*
 * Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
 *
 * See doc/protocol.md and https://github.com/trevp/noise/blob/master/noise.md for more info
 */

#ifndef NOISE_H
#define NOISE_H

#include "../crypto/curve25519.h"
#include "../crypto/chacha20poly1305.h"
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/atomic.h>
#include <linux/rwsem.h>
#include <linux/mutex.h>
#include <linux/jiffies.h>

enum noise_lengths {
	NOISE_PUBLIC_KEY_LEN = CURVE25519_POINT_SIZE,
	NOISE_SYMMETRIC_KEY_LEN = CHACHA20POLY1305_KEYLEN,
	NOISE_NONCE_LEN = CHACHA20POLY1305_NONCELEN,
	NOISE_TIMESTAMP_LEN = 12,
	NOISE_AUTHTAG_LEN = CHACHA20POLY1305_AUTHTAGLEN,
	NOISE_HASH_LEN = 32
};

enum wireguard_timeouts {
	REKEY_AFTER_MESSAGES = U64_MAX - 0xffff,
	REJECT_AFTER_MESSAGES = U64_MAX - 0xf,
	REKEY_TIMEOUT = 5 * HZ,
	REKEY_AFTER_TIME = 120 * HZ,
	REJECT_AFTER_TIME = REKEY_AFTER_TIME + (6 * REKEY_TIMEOUT)
};

union noise_counter {
	struct {
		u64 counter;
		unsigned long backtrack;
		spinlock_t lock;
	} receive;
	atomic64_t counter;
};

struct noise_symmetric_key {
	bool is_valid;
	u8 key[NOISE_SYMMETRIC_KEY_LEN];
	unsigned long birthdate;
	u16 index;
	union noise_counter counter;
};

struct noise_handshake {
	bool is_valid;
	u8 latest_timestamp[NOISE_TIMESTAMP_LEN];

	u8 static_public[NOISE_PUBLIC_KEY_LEN];
	u8 static_private[NOISE_PUBLIC_KEY_LEN];

	u8 ephemeral_public[NOISE_PUBLIC_KEY_LEN];
	u8 ephemeral_private[NOISE_PUBLIC_KEY_LEN];

	u8 remote_static[NOISE_PUBLIC_KEY_LEN]; /* Immutable after noise_peer_init; not protected by mutex. */
	u8 remote_ephemeral[NOISE_PUBLIC_KEY_LEN];

	u8 hash[NOISE_HASH_LEN];
	u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN];

	struct noise_symmetric_key key;
	struct rw_semaphore lock;
};

struct noise_session_keys {
	struct noise_symmetric_key *sending;
	struct noise_symmetric_key *receiving;
	struct noise_symmetric_key *previous_sending;
	struct noise_symmetric_key *previous_receiving;
	bool send_with_previous;
	bool i_am_the_initator;
	struct rcu_head rcu;
};

struct noise_peer {
	struct noise_handshake handshake;
	struct noise_session_keys __rcu *session;
	struct mutex key_update_lock;
};

#define noise_encrypted_len(plain_len) (plain_len + NOISE_AUTHTAG_LEN)

enum noise_message_type {
	NOISE_MESSAGE_INVALID = 0,
	NOISE_MESSAGE_HANDSHAKE_INITIATION = 1,
	NOISE_MESSAGE_HANDSHAKE_RESPONSE = 2,
	NOISE_MESSAGE_DATA = 3,
	NOISE_MESSAGE_TOTAL = 4
};

struct noise_message_header {
	u8 type;
} __attribute__((packed));

struct noise_message_handshake_initiation {
	struct noise_message_header header;
	u16 sender_index;
	u8 unencrypted_ephemeral[NOISE_PUBLIC_KEY_LEN];
	u8 encrypted_static[noise_encrypted_len(NOISE_PUBLIC_KEY_LEN)];
	u8 encrypted_timestamp[noise_encrypted_len(NOISE_TIMESTAMP_LEN)];
} __attribute__((packed));

struct noise_message_handshake_response {
	struct noise_message_header header;
	u16 sender_index;
	u16 receiver_index;
	u8 encrypted_ephemeral[noise_encrypted_len(NOISE_PUBLIC_KEY_LEN)];
	u8 encrypted_nothing[noise_encrypted_len(0)];
} __attribute__((packed));

struct noise_message_data {
	struct noise_message_header header;
	u16 key_idx;
	u64 counter;
	u8 encrypted_data[];
} __attribute__((packed));
#define noise_message_data_len(plain_len) (noise_encrypted_len(plain_len) + sizeof(struct noise_message_data))

enum noise_message_alignments {
	NOISE_MESSAGE_DATA_TARGET_OFFSET = sizeof(struct noise_message_data),
	NOISE_MESSAGE_DATA_TARGET_OPTIMAL_ALIGNMENT = 32 /* Per intel AVX recommendations */
};

struct wireguard_device;
struct wireguard_peer;

/* Internal API: */
bool noise_message_encrypt(struct noise_message_data *dst, const u8 *src_plaintext, size_t src_len, struct noise_symmetric_key *key);
bool noise_message_decrypt(u8 *dst_plaintext, const struct noise_message_data *src, size_t src_len, struct noise_symmetric_key *key);
bool noise_handshake_encrypt(u8 *dst_ciphertext, const u8 *src_plaintext, size_t src_len, struct noise_symmetric_key *key, u8 hash[NOISE_HASH_LEN]);
bool noise_handshake_decrypt(u8 *dst_plaintext, const u8 *src_ciphertext, size_t src_len, struct noise_symmetric_key *key, u8 hash[NOISE_HASH_LEN]);
void noise_handshake_nocrypt(u8 *dst, const u8 *src, size_t src_len, u8 hash[NOISE_HASH_LEN]);
void noise_symmetric_key_init(struct noise_symmetric_key *key, const u8 responder_static[NOISE_PUBLIC_KEY_LEN]);
void noise_handshake_clear(struct noise_handshake *handshake);
bool noise_mix_dh(struct noise_symmetric_key *key, u8 chaining_key[NOISE_SYMMETRIC_KEY_LEN], const u8 private[NOISE_PUBLIC_KEY_LEN], const u8 public[NOISE_PUBLIC_KEY_LEN]);

/* External API: */
bool noise_handshake_create_initiation(struct noise_message_handshake_initiation *dst, struct noise_peer *peer, u16 self_index);
struct wireguard_peer *noise_handshake_consume_initiation(struct noise_message_handshake_initiation *src, struct wireguard_device *wg);
bool noise_handshake_create_response(struct noise_message_handshake_response *dst, struct noise_peer *peer, u16 self_index);
struct wireguard_peer *noise_handshake_consume_response(struct noise_message_handshake_response *src, struct wireguard_device *wg);
bool noise_handshake_begin_session(struct noise_peer *peer, bool i_am_the_initator);
bool noise_message_create_data(struct noise_message_data *dst, const u8 *src, size_t src_len, struct noise_peer *peer);
struct wireguard_peer *noise_message_consume_data(u8 *dst, const struct noise_message_data *src, size_t src_len, struct wireguard_device *wg, bool *new_key);
void noise_peer_init(struct noise_peer *peer, const u8 my_private_key[NOISE_PUBLIC_KEY_LEN], const u8 peer_public_key[NOISE_PUBLIC_KEY_LEN]);
void noise_peer_set_private_key(struct noise_peer *peer, const u8 my_private_key[NOISE_PUBLIC_KEY_LEN]);
void noise_peer_clear(struct noise_peer *peer);
void noise_peer_free_all_unlocked(struct noise_peer *peer);
static inline u8 noise_message_determine_type(void *src, size_t src_len)
{
	struct noise_message_header *header = src;
	if (unlikely(src_len < sizeof(struct noise_message_header)))
		return NOISE_MESSAGE_INVALID;
	if (unlikely(header->type >= NOISE_MESSAGE_TOTAL))
		return NOISE_MESSAGE_INVALID;
	return header->type;
}

#endif
