/* Copyright 2015 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */

#include "../wireguard.h"
#include "noise.h"
#include <linux/rcupdate.h>
#include <linux/slab.h>

void noise_peer_init(struct noise_peer *peer, const u8 my_private_key[NOISE_PUBLIC_KEY_LEN], const u8 peer_public_key[NOISE_PUBLIC_KEY_LEN])
{
	memset(peer, 0, sizeof(struct noise_peer));
	init_rwsem(&peer->handshake.lock);
	mutex_init(&peer->key_update_lock);
	memcpy(peer->handshake.remote_static, peer_public_key, NOISE_PUBLIC_KEY_LEN);
	noise_peer_set_private_key(peer, my_private_key);
}

void noise_peer_set_private_key(struct noise_peer *peer, const u8 my_private_key[NOISE_PUBLIC_KEY_LEN])
{
	down_write(&peer->handshake.lock);
	memcpy(peer->handshake.static_private, my_private_key, NOISE_PUBLIC_KEY_LEN);
	curve25519_generate_public(peer->handshake.static_public, peer->handshake.static_private);
	up_write(&peer->handshake.lock);
}

static inline void free_session(struct noise_session_keys *session)
{
	if (!session)
		return;
	if (session->previous_receiving) {
		memzero_explicit(session->previous_receiving, sizeof(struct noise_symmetric_key));
		kfree(session->previous_receiving);
	}
	if (session->previous_sending) {
		memzero_explicit(session->previous_sending, sizeof(struct noise_symmetric_key));
		kfree(session->previous_sending);
	}
	if (session->receiving) {
		memzero_explicit(session->receiving, sizeof(struct noise_symmetric_key));
		kfree(session->receiving);
	}
	if (session->sending) {
		memzero_explicit(session->sending, sizeof(struct noise_symmetric_key));
		kfree(session->sending);
	}
	kfree(session);
	pr_debug("Old noise session cleared completely and freed\n");
}

static void clear_entire_session(struct rcu_head *rcu)
{
	free_session(container_of(rcu, struct noise_session_keys, rcu));
}

void noise_peer_clear(struct noise_peer *peer)
{
	struct noise_session_keys *old_session;

	noise_handshake_clear(&peer->handshake);

	mutex_lock(&peer->key_update_lock);
	old_session = rcu_dereference_protected(peer->session, lockdep_is_held(&peer->key_update_lock));
	rcu_assign_pointer(peer->session, NULL);
	mutex_unlock(&peer->key_update_lock);
	if (old_session)
		call_rcu(&old_session->rcu, clear_entire_session);
}

void noise_peer_free_all_unlocked(struct noise_peer *peer)
{
	rcu_read_lock();
	free_session(rcu_dereference(peer->session));
	rcu_read_unlock();
	memzero_explicit(peer, sizeof(struct noise_peer));
}
